import ee
import os
import re
from rlcms.composites import Composite
from rlcms.utils import check_exists
import argparse
import json

def main():
    ee.Initialize()
    parser = argparse.ArgumentParser(
    description="Create a Composite from one or multiple datasets",
    usage = "composite -a aoi/fc/path -d Landsat8 -s 2020-01-01 -e 2020-12-31 -o output/path --settings path/to/settings/file.txt "
    )
    
    parser.add_argument(
    "-a",
    "--aoi",
    type=str,
    required=True,
    help="The asset path to an aoi or reference polygon dataset"
    )
    
    parser.add_argument(
    "-d",
    "--data",
    type=str, 
    nargs='+',
    required=True,
    help="Dataset(s) to composite"
    )
    
    parser.add_argument(
    "-s",
    "--start",
    type=str,
    required=True,
    help="start date (yyyy-mm-dd)"
    )

    parser.add_argument(
    "-e",
    "--end",
    type=str,
    required=True,
    help="end date (yyyy-mm-dd)"
    )

    parser.add_argument(
    "-o",
    "--output",
    type=str,
    required=True,
    help="The full asset path for export"
    )

    parser.add_argument(
    "--settings",
    type=str,
    required=True,
    help="settings .txt file"
    )
    
    parser.add_argument(
    "--scale",
    type=int,
    required=True,
    help="Scale of output composite."
    )
    
    parser.add_argument(
    "--crs",
    type=str,
    required=False,
    help="CRS string in format of EPSG:xxxxx. Defaults to EPSG:4326"
    )
    
    parser.add_argument(
    "--dry_run",
    dest="dry_run",
    action="store_true",
    help="goes through checks and prints output asset path but does not export",
    )
    
    args = parser.parse_args()
    
    aoi_path = args.aoi
    data = args.data
    start = args.start
    end = args.end
    output = args.output
    scale = args.scale
    crs = args.crs
    settings_f = args.settings
    dry_run = args.dry_run
    
    output_folder = os.path.dirname(output)
   
    # parse txt file into settings dict that we'll pass to composite()
    with open(settings_f) as f:
        settings = json.load(f)
    
    # check output doesn't already exist (GEE by default prohibits overwrites)
    assert check_exists(output), f"Image already exsits: {output}"
    # check inputs exist
    assert check_exists(aoi_path) == 0, f"Check aoi exists: {aoi_path}"
    assert check_exists(output_folder) == 0, f"Check output folder exists: {output_folder}"
    # ensure start and end dates will work
    for date in [start,end]:
        r = re.compile('.{4}-.{2}-.{2}')
        if len(date) == 10:
            if r.match(date):
                pass
            else:
                raise ValueError(f"date string(s) don't match required format. Got: {date}")
        else:
            raise ValueError(f"date string(s) don't match required format. Got: {date}")
    
    aoi = ee.FeatureCollection(aoi_path)
    
    # multiple datasets requested, combining multiple composites
    if len(data) > 1: 
        composite_list = ([Composite(dataset=d,
                        region=aoi,
                        start_date=start,
                        end_date=end,
                        **settings)
                        # prefix dataset name to every band, if data is an asset path we swap / for _
                        .regexpRename('^', f"{d.replace('/','_')}_") 
                            for d in data])
        
        img = ee.Image.cat(composite_list)
    
    # only one dataset requested
    else:
        img = Composite(dataset=data[0],
                        region=aoi,
                        start_date=start,
                        end_date=end,
                        **settings)
    
    if dry_run:
        print(f"would export: {output}")
    
    else:
        if crs == None:
            task = ee.batch.Export.image.toAsset(image=img.image,
                                        description=os.path.basename(output),
                                        assetId=output,
                                        region=aoi.geometry(),
                                        scale=scale,
                                        maxPixels=1e12)
        else:
            task = ee.batch.Export.image.toAsset(image=img.image,
                                        description=os.path.basename(output),
                                        assetId=output,
                                        region=aoi.geometry(),
                                        scale=scale,
                                        maxPixels=1e12,
                                        crs=crs)
        task.start()
        print(f"Export started (Asset): {output}") 

if __name__ == "__main__":
    main()


#%%
import os
import ee
import argparse
import numpy as np
import rlcms.sampling as sampling
from rlcms.utils import check_exists, exportTableToAsset, exportTableToDrive

def main():
    ee.Initialize()

    parser = argparse.ArgumentParser(
    description="Generate Random Sample Points From an ee.Image, Formatted for Collect Earth Online",
    usage = "sample_pts -im input/path/to/image -band LANDCOVER -o output/path --n_points 100 --to_drive"
    )
    
    parser.add_argument(
    "-im",
    "--input_image",
    type=str,
    required=True,
    help="asset path to image you are sampling from"
    )
    
    # added this required arg for greater flexibility. default behavior of .stratifiedSample() is to use image's first band
    # most LC images will only have one band but in case the img doesn't this ensures correct behavior..
    parser.add_argument(
    "-band",
    "--class_band",
    type=str,
    required=True,
    help="class band name to use for stratification"
    )

    parser.add_argument(
        "-s",
        "--scale",
        type=int,
        required=True,
        help="scale to conduct sampling (meters)"
        )
    
    parser.add_argument(
    "-oa",
    "--output_asset",
    type=str,
    required=False,
    help="The output GEE asset path for export."
    )
    
    parser.add_argument(
    "-od",
    "--output_drive",
    type=str,
    required=False,
    help="The output google drive path for export."
    )
    parser.add_argument(
    "--n_points",
    type=int,
    required=False,
    help="Number of points per class. Default: 100"
    )

    parser.add_argument(
    '--class_values', 
    type=int, 
    nargs='+',
    required=False,
    help="list of unique LANDCOVER values in input Feature Collection"
    )

    parser.add_argument(
    "--class_points",
    type=int,
    nargs='+',
    required=False,
    help="number of samples to collect per class"
    )
    
    parser.add_argument(
    "-r",
    "--reshuffle",
    dest="reshuffle",
    action="store_true",
    help="randomizes seed used in all functions",
    )

    parser.add_argument(
    "-d",
    "--dry_run",
    dest="dry_run",
    action="store_true",
    help="goes through checks and prints output asset path but does not export.",
    )

    args = parser.parse_args()
    
    input_path = args.input_image
    class_band = args.class_band
    scale = args.scale
    output_asset = args.output_asset
    output_drive = args.output_drive
    n_points = args.n_points
    class_values = args.class_values
    class_points = args.class_points
    reshuffle = args.reshuffle
    dry_run = args.dry_run

    # perform checks
    assert check_exists(input_path) == 0, f"Check input FeatureCollection exists: {input_path}"
    # GEE won't make parents for you
    if output_asset:
        output_asset_folder = os.path.dirname(output_asset)
        assert check_exists(output_asset_folder) == 0, f"Check GEE output asset's folder exists: {output_asset_folder}"
    
    # value checks if class_values and class_points args are both provided
    if ((class_values != None) and (class_points != None)):
        
        # class_values and class_points must be equal length
        if len(class_values) != len(class_points):
            print(f"Error: class_points and class_values are of unequal length: {class_values} {class_points}")
            exit()
        # class_values and class_points provided so we'll override n_points, 
        # but ee.Image.stratifiedSample() still requires it so we set to default
        n_points=100
    
    # if only one is provided, error 
    elif (class_values != None and class_points == None) or (class_values == None and class_points != None):
        print(f"Error: class_values and class_points args are codependent, provide both or neither. class_values:{class_values}, class_points:{class_points}")
    
    # if neither class_values nor class_points provided, n_points must be provided, otherwise we set a default n_points value 
    else:
        if n_points != None:
            pass
        else:
            n_points=100
            print(f"Warning: Defaulting to equal allocation of default n: {n_points}. Set n_points or class_values and class_points to control sample allocation.")
            
    
    img = ee.Image(input_path)
    bbox = img.geometry().bounds() # region
    # default seed is set, will re-randomize seed if reshuffle==True
    seed=90210
    if reshuffle:
        np.random.RandomState()
        seed = np.random.randint(low=1,high=1e6)
        print(f"reshuffled new seed: {seed}")
    
    
    samples = sampling.strat_sample(img=img,
                                    class_band=class_band,
                                    region=bbox,
                                    scale=scale, # 10, hard coded set at top of script, can make this a user arg for greater flexibility
                                    seed=seed, 
                                    n_points=n_points,
                                    class_values=class_values,
                                    class_points=class_points)

   
    selectors = 'LON,LAT,PLOTID,SAMPLEID,'+class_band
    
    # to GEE Asset only
    if output_asset!=None and output_drive==None:
        desc = os.path.basename(output_asset)+'-Asset'
        if dry_run:
            print(f"would export (Asset): {output_asset}")
            exit()
        else:
            exportTableToAsset(samples,desc,output_asset)
    
    # to Google Drive only
    elif output_asset==None and output_drive!=None:
        # user might not provide a GEE asset conforming path as output if toAsset not their intent
        drive_folder = os.path.dirname(output_drive)
        drive_basename = os.path.basename(output_drive)
        drive_desc = drive_basename+'-Drive'
        if dry_run:
            print(f"would export (Drive): {output_drive}")
            exit()
        else:
            exportTableToDrive(samples,drive_desc,drive_folder,drive_basename,selectors)
    
    # export both to GEE Asset and Google Drive
    elif output_asset!=None and output_drive!=None: 
        asset_desc = os.path.basename(output_asset)+'-Asset'
        drive_folder = os.path.dirname(output_drive)
        drive_basename = os.path.basename(output_drive)
        drive_desc = drive_basename+'-Drive'
        if dry_run:
            print(f"would export (Asset): {output_asset}")
            print(f"would export (Drive): {output_drive}")
            exit()
        else:
            exportTableToAsset(samples,asset_desc,output_asset)
            exportTableToDrive(samples,drive_desc,drive_folder,drive_basename,selectors)
    # neither output paths specified
    else:
        raise RuntimeError(f"No output paths provided, user must provide at least one of the following two arguments:\n -oa/--output_asset  -od/--output_drive")        
if __name__ == "__main__":
    main()



import ee
import os
import pandas as pd
from rlcms.utils import export_img_to_asset, export_image_to_drive
from ee.ee_exception import EEException
import subprocess

class Primitives:
    def __init__(self,
                 inputs=None,
                 training=None,
                 class_name=None,
                 asset_id=None):
        """
        Construct a Primitives ensemble, provided an input ee.Image stack containing feature bands and a training point FeatureCollection
        
        Args:
            inputs (str|ee.Image): input image stack
            training (str|ee.FeatureCollection): training data
            class_name (str): class property containing class labels (i.e. 1, 2, 3), currently only 'LANDCOVER' is supported
            asset_id (str): Optional, GEE asset path to pre-existing Primitives ee.ImageCollection. Useful for exporting intermediary output approach
        
        Returns: 
            Primitives object
        """
        
        # TODO: perform some checks and error handling for training point label formatting 
        def pre_format_pts(pts,class_name):
            # two things: 
            # 1. need to ensure that 'LANDCOVER' is the class_name, if not, set new property in each feature using that
            if class_name != 'LANDCOVER':
                # set user's class_name to 'LANDCOVER' so we don't need to pass custom class_name around to every function
                pts = pts.map(lambda p: p.set('LANDCOVER',p.get(class_name)))
            else:
                # check 'LANDCOVER' is a property in collection
                assert 'LANDCOVER' in pts.first().propertyNames().getInfo(), "'LANDCOVER' is not a property in the collection"
            # 2. ensure that value of 'LANDCOVER' is integer before converting to string
            def to_int(p):
                val = p.get('LANDCOVER')
                # is 'LANDCOVER' always going to be string? can we convert it to string without first knowing if its String or Number?
                int_val = ee.Number(val).round()
                return p.set('LANDCOVER',int_val)
            return pts.map(to_int)
        
        def format_pts(pts):
            """Turn a FC of training points containing full LC typology into a list of primitive point FCs, 
                    one point FC for each LC primitive"""
            # create sets of binary training pts for each class represented in the full training pts collection
            labels = ee.FeatureCollection(pts).aggregate_array('LANDCOVER').distinct().sort()
            def binaryPts(l):
                # create prim and non prim sets with filters, reset prim to 1, non-prim to 0
                prim = pts.filter(ee.Filter.eq('LANDCOVER',l)).map(lambda f: f.set('PRIM',1))
                non_prim = pts.filter(ee.Filter.neq('LANDCOVER',l)).map(lambda f: f.set('PRIM',0))
                return ee.FeatureCollection(prim).merge(non_prim)
            list_of_prim_pts = ee.List(labels).map(binaryPts)
            return list_of_prim_pts

        def gettop20(dict):
            # if total input features count < 20, take them all, otherwise take top 20 most important
            dict = ee.Dictionary(dict)
            values = dict.values().sort()
            cutoff = ee.Algorithms.If(values.size().gte(20),-20,values.size().multiply(-1))
            def kv_return(key,passedObj):
                passedObj = ee.List(passedObj)
                val = ee.Number(dict.get(key))
                retObj = ee.Algorithms.If(val.gte(cutoff),passedObj.add(key),passedObj)
                return retObj
            newl = dict.keys().iterate(kv_return,ee.List([]))
            return newl
        
        def RFprim(training_pts,input_stack):
            """Train and apply RF Probability classifier on a Primitive"""
            inputs = ee.Image(input_stack)
            samples = ee.FeatureCollection(training_pts)
            
            class_value = ee.Number(ee.Feature(samples.sort('PRIM',False).first()).get('LANDCOVER')) #get LC numeric value for the given primitive (i.e. 'PRIM':1, 'LANDCOVER':6) then map to its class label (i.e. 6: 'Water')
            
            # can experiment with classifier params for model performance
            classifier = ee.Classifier.smileRandomForest(
            numberOfTrees=100, 
            minLeafPopulation=1, 
            bagFraction=0.7, 
            seed=51515).setOutputMode('PROBABILITY')
            
            # train model with all features
            model = classifier.train(features=samples, 
                                    classProperty='PRIM', 
                                    inputProperties=inputs.bandNames() 
                                    )
            
            # store for model performance exploration
            oob_all = ee.Dictionary(model.explain()).get('outOfBagErrorEstimate')
            importance_all = ee.Dictionary(model.explain()).get('importance')
            
            # retrieve top 20 most important features
            top20 = gettop20(importance_all)
            
            # re-train model with top20 important features
            model = classifier.train(features=samples, 
                                    classProperty='PRIM', 
                                    inputProperties=top20
                                    )
            
            oob_top20 = ee.Dictionary(model.explain()).get('outOfBagErrorEstimate')
            importance_top20 = ee.Dictionary(model.explain()).get('importance')
            schema = ee.List(ee.Classifier(model).schema())
            output = (ee.Image(inputs)
                      .classify(model,'Probability')
                      # lists and dictionaries do not propagate thru as properties on Batch exported Images
                      .set('Primitive',class_value,
                           'importance',importance_top20, 
                           'schema',schema, 
                           'model',model,
                           'oobError',oob_top20, 
                           ))
            return output

        def primitives_to_collection(input_stack,
                                     training_pts,
                                     class_name):
            """
            Create LC Primitive image for each LC class in training points

            args:
                input_stack (ee.Image): of all covariates and predictor
                training_pts (ee.FeatureCollection): training pts containing full LC typology
                class_name (str): property name in training points containing model classes
            
            returns:
                ee.ImageCollection of Primitive ee.Images
            """

            input_stack = ee.Image(input_stack)
            training_pts = ee.FeatureCollection(training_pts)
            
            # list of distinct LANDCOVER values
            labels = training_pts.aggregate_array(class_name).distinct().sort().getInfo() # .sort() should fix Prims exporting out of order (i.e. 2,3,4,7,6)

            # converting to index of the list of distinct LANDCOVER primtive FC's (prim_pts below)
            indices = list(range(len(labels))) # handles dynamic land cover strata

            prim_list = []
            for i in indices: # running one LC class at a time
                prim_pts = ee.FeatureCollection(ee.List(format_pts(training_pts)).get(i)) # format training pts to 1/0 prim format
                img = RFprim(prim_pts,input_stack) # run RF primitive model, get output image and metrics
                prim_list.append(img)
            
            return ee.ImageCollection.fromImages(prim_list)
        
        # you can construct Primitives object from a pre-existing Primitives ImgColl
        if asset_id != None:
            try:
                primitives = ee.ImageCollection(asset_id)
                self.collection = primitives
                self.region = primitives.first().geometry().getInfo()
                self.training_data = None
            except: 
                raise(EEException)
        else:
            primitives = primitives_to_collection(inputs,training,class_name)
            self.collection = primitives
            self.region = ee.Image(inputs).geometry().getInfo()
            self.training_data = ee.FeatureCollection(training)
    
    def assemble_max_probability(self, remap_to:list=None):
        """
        Perform pixel-wise max probability assemblage method. At each pixel, the primitive with highest probability is returned in the assemblage image.
        If your desired land cover typology does not start with 1 and/or skips values, you must specify a remap_to list that matches the desired output land cover typology.
        e.g. remap_to=[1,2,3,6,11,12,13]            
        
        Args:
            remap_to: list, default=None, list of integers matching the desired output land cover typology

        Returns: ee.Image                            
        """
        def max_prob(image):
            
            maxProbClassification = (image.toArray()
                                    .arrayArgmax()
                                    .arrayFlatten([['classification']])
                                    .rename('classification')
                                    )
            return maxProbClassification
        
        image  = self.collection.toBands()
        max_probability = max_prob(image)
        output = max_probability.add(1) # shift values from 0-n to 1-n, where n = bands
        
        if remap_to != None:
            prims_count = self.collection.size().getInfo()
            remap_from = list(range(1,prims_count+1)) # values without a remap are 1 to n
            if len(remap_from) != len(remap_to):
                raise ValueError("remap_to must be the same length as the number of primitives in the collection", 
                                 f"remap_from: {len(remap_from)}, remap_to: {len(remap_to)}") 
            output = output.remap(remap_from, remap_to)
        else:
            pass
        return output.rename('LANDCOVER')
        
    def export_metrics(self,metrics_path):
            """
            Parse variable importance and OOB Error estimate from trained model, output to local files respectively
            Currently only works for Primitives objects in memory (not loaded from pre-existing ImgColl)
            """
            # Var Importance to csv file
            imgColl = self.collection
            to_list = ee.List(imgColl.toList(imgColl.size()))
            for i in list(range(imgColl.size().getInfo())):
                img = ee.Image(to_list.get(i))
                prim_value = str(img.get('Primitive').getInfo())
                
                # Variable Importance to .csv
                dct = ee.Dictionary(img.get('importance')).getInfo()
                _list = dct.values()
                idx = dct.keys()
                df = pd.DataFrame(_list, index = idx)
                df.to_csv(os.path.join(metrics_path,f"varImportancePrimitive{prim_value}.csv"))
                
                # OOB error to .txt file
                oob = img.get('oobError')
                with open(os.path.join(metrics_path,f'oobErrorPrimitive{prim_value}.txt'),mode='w') as f:
                    f.write(ee.String(ee.Number(oob).format()).getInfo())
                    f.close()
    
    def export_to_asset(self,
                        collection_assetId=None,
                        scale=None,
                        crs=None,
                        crsTransform=None,
                        maxPixels=None,
                        **kwargs):
        """
        Export Primitives to Asset as an ImageCollection
        
        Args:
            collection_assetId (str): output ImageCollection asset path 
            scale (int): export scale
            crs (str): export CRS ('EPSG:4326')
            crsTransform (list): export CRS Transform
            maxPixels (int): max Pixels
        
        Returns: 
            None, Submits all Export Image tasks for Primitive collection
        """
        
        # make the empty IC
        command = f'earthengine create collection {collection_assetId}'
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        exit_code = process.returncode
        if exit_code == 0:
            print(f"Created empty Primitives ImageCollection: {collection_assetId}")

        else:
            raise RuntimeError(f"Process failed with exit code {exit_code}",
                            f"Error message: {stdout.decode()}")
        
        # os.popen(f"earthengine create collection {collection_assetId}").read()
        prims_count = self.collection.size().getInfo()
        prims_list = ee.ImageCollection(self.collection).toList(prims_count)
        aoi = ee.Image(prims_list.get(0)).geometry()
        for i in list(range(prims_count)):
            prim = ee.Image(prims_list.get(i))
            desc = f"Primitive{str(ee.Image(prim).get('Primitive').getInfo())}" # this would need to be defined in the Prims img for-loop
            asset_id = f'{collection_assetId}/{desc}'
            export_img_to_asset(image=prim,
                                description=desc,
                                assetId=asset_id,
                                region=aoi,
                                scale=scale,
                                crs=crs,
                                crsTransform=None,
                                maxPixels=1e13)
            
        return
    
    def export_to_drive(self,
                        description,
                        folder,
                        fileNamePrefix,
                        dimensions=None,
                        region=None,
                        scale=None,
                        crs=None,
                        crsTransform=None,
                        maxPixels=None,
                        shardSize=None,
                        fileDimensions=None,
                        skipEmptyTiles=None,
                        fileFormat=None,
                        formatOptions=None,
                        **kwargs,):
        """
        Export Primitives to Drive as a Multi-band GeoTiff
        
        See rlcms.utils.export_img_to_drive() docs for Args
        
        Returns: 
            None, Submits all Export Image tasks for Primitive collection
        """
        
        prim_img = self.collection.toBands()
        export_image_to_drive(image=prim_img,
                                description=description,
                                folder=folder,
                                fileNamePrefix=fileNamePrefix,
                                dimensions=dimensions,
                                region=region,
                                scale=scale,
                                crs=crs,
                                crsTransform=crsTransform,
                                maxPixels=maxPixels,
                                shardSize=shardSize,
                                fileDimensions=fileDimensions,
                                skipEmptyTiles=skipEmptyTiles,
                                fileFormat=fileFormat,
                                formatOptions=formatOptions,
                                **kwargs,)
        return

    
    


import ee
import json
ee.Initialize()

def parse_settings(settings):
    if isinstance(settings,str): # file path string
        # print('is a file on their computer')
        # print(settings)
        try:
            with open(settings,mode='r') as f:
                data = f.read()
            # reconstructing the data as a dictionary
            settings = json.loads(data)
        except:
            raise RuntimeError(f'Could not parse file path: {settings}')  
    else: 
        raise TypeError(f'expects a str, got: {type(settings)}')
    
    return settings

def check_exists(ee_path:str):
    try:
        ee.data.getAsset(ee_path)
        return 0 # does exist returns 0/False
    except ee.ee_exception.EEException:
        return 1 # doesn't exist returns 1/True

def export_image_to_drive(
    image,
    description="myExportImageTask",
    folder=None,
    fileNamePrefix=None,
    dimensions=None,
    region=None,
    scale=None,
    crs=None,
    crsTransform=None,
    maxPixels=None,
    shardSize=None,
    fileDimensions=None,
    skipEmptyTiles=None,
    fileFormat=None,
    formatOptions=None,
    **kwargs,
):
    """Creates a batch task to export an Image as a raster to Google Drive.

    Args:
        image: The image to be exported.
        description: Human-readable name of the task.
        folder: The name of a unique folder in your Drive account to
            export into. Defaults to the root of the drive.
        fileNamePrefix: The Google Drive filename for the export.
            Defaults to the name of the task.
        dimensions: The dimensions of the exported image. Takes either a
            single positive integer as the maximum dimension or "WIDTHxHEIGHT"
            where WIDTH and HEIGHT are each positive integers.
        region: The lon,lat coordinates for a LinearRing or Polygon
            specifying the region to export. Can be specified as a nested
            lists of numbers or a serialized string. Defaults to the image's
            region.
        scale: The resolution in meters per pixel. Defaults to the
            native resolution of the image assset unless a crsTransform
            is specified.
        crs: The coordinate reference system of the exported image's
            projection. Defaults to the image's default projection.
        crsTransform: A comma-separated string of 6 numbers describing
            the affine transform of the coordinate reference system of the
            exported image's projection, in the order: xScale, xShearing,
            xTranslation, yShearing, yScale and yTranslation. Defaults to
            the image's native CRS transform.
        maxPixels: The maximum allowed number of pixels in the exported
            image. The task will fail if the exported region covers more
            pixels in the specified projection. Defaults to 100,000,000.
        shardSize: Size in pixels of the tiles in which this image will be
            computed. Defaults to 256.
        fileDimensions: The dimensions in pixels of each image file, if the
            image is too large to fit in a single file. May specify a
            single number to indicate a square shape, or a tuple of two
            dimensions to indicate (width,height). Note that the image will
            still be clipped to the overall image dimensions. Must be a
            multiple of shardSize.
        skipEmptyTiles: If true, skip writing empty (i.e. fully-masked)
            image tiles. Defaults to false.
        fileFormat: The string file format to which the image is exported.
            Currently only 'GeoTIFF' and 'TFRecord' are supported, defaults to
            'GeoTIFF'.
        formatOptions: A dictionary of string keys to format specific options.
        **kwargs: Holds other keyword arguments that may have been deprecated
            such as 'crs_transform', 'driveFolder', and 'driveFileNamePrefix'.
    """

    if not isinstance(image, ee.Image):
        raise ValueError("Input image must be an instance of ee.Image")

    task = ee.batch.Export.image.toDrive(
        image,
        description,
        folder,
        fileNamePrefix,
        dimensions,
        region,
        scale,
        crs,
        crsTransform,
        maxPixels,
        shardSize,
        fileDimensions,
        skipEmptyTiles,
        fileFormat,
        formatOptions,
        **kwargs,
    )
    task.start()
    print(f"Export Started (Drive): {fileNamePrefix}")

def export_img_to_asset(image,
    description="myExportImageTask",
    assetId=None,
    pyramidingPolicy=None,
    dimensions=None,
    region=None,
    scale=None,
    crs=None,
    crsTransform=None,
    maxPixels=None,
    **kwargs):
    """Creates a task to export an EE Image to an EE Asset.

    Args:
        image: The image to be exported.
        description: Human-readable name of the task.
        assetId: The destination asset ID.
        pyramidingPolicy: The pyramiding policy to apply to each band in the
            image, a dictionary keyed by band name. Values must be
            one of: "mean", "sample", "min", "max", or "mode".
            Defaults to "mean". A special key, ".default", may be used to
            change the default for all bands.
        dimensions: The dimensions of the exported image. Takes either a
            single positive integer as the maximum dimension or "WIDTHxHEIGHT"
            where WIDTH and HEIGHT are each positive integers.
        region: The lon,lat coordinates for a LinearRing or Polygon
            specifying the region to export. Can be specified as a nested
            lists of numbers or a serialized string. Defaults to the image's
            region.
        scale: The resolution in meters per pixel. Defaults to the
            native resolution of the image assset unless a crsTransform
            is specified.
        crs: The coordinate reference system of the exported image's
            projection. Defaults to the image's default projection.
        crsTransform: A comma-separated string of 6 numbers describing
            the affine transform of the coordinate reference system of the
            exported image's projection, in the order: xScale, xShearing,
            xTranslation, yShearing, yScale and yTranslation. Defaults to
            the image's native CRS transform.
        maxPixels: The maximum allowed number of pixels in the exported
            image. The task will fail if the exported region covers more
            pixels in the specified projection. Defaults to 100,000,000.
        **kwargs: Holds other keyword arguments that may have been deprecated
            such as 'crs_transform'.
    """

    if isinstance(image, ee.Image) or isinstance(image, ee.image.Image):
        pass
    else:
        raise ValueError("Input image must be an instance of ee.Image")

    if isinstance(assetId, str):
        if assetId.startswith("users/") or assetId.startswith("projects/"):
            pass
        else:
            assert check_exists(assetId) == 0, f"{assetId} not a valid asset path"
            # assetId = f"{ee_user_id()}/{assetId}"

    task = ee.batch.Export.image.toAsset(
        image,
        description,
        assetId,
        pyramidingPolicy,
        dimensions,
        region,
        scale,
        crs,
        crsTransform,
        maxPixels,
        **kwargs,
    )
    task.start()
    print(f"Export Started (Asset): {assetId}")

def exportTableToAsset(collection:ee.FeatureCollection,description:str,asset_id:str):
    """Export FeatureCollection to GEE Asset"""
    if check_exists(asset_id) == 1:
        task = ee.batch.Export.table.toAsset(
            collection=collection,
            description=description,
            assetId=asset_id,
            )
        task.start()
        print(f'Export started (Asset): {asset_id}')
    else:
        print(f"{asset_id} already exists")
    
    return

def exportTableToDrive(collection:ee.FeatureCollection,description:str,folder:str,file_name_prefix:str,selectors:str):
    """export FeatureCollection to Google Drive"""
    task = ee.batch.Export.table.toDrive(
        collection=collection, 
        description=description, 
        folder=folder,
        fileNamePrefix=file_name_prefix,
        selectors=selectors)
    task.start()
    print(f'Export started (Drive): {folder}/{file_name_prefix}')
    return


import os
import ee
import argparse
from rlcms.primitives import Primitives
from rlcms.utils import check_exists, export_img_to_asset
    
def main():
    ee.Initialize()
    
    parser = argparse.ArgumentParser(
    description="Generate Single Land Cover Image From Land Cover Primitives Image Collection",
    usage = "generate_LC -i path/to/input_primitive_collection -o output/path/to/landcover_image"
    )
    
    parser.add_argument(
    "-i",
    "--input",
    type=str,
    required=True,
    help="GEE asset path to input Primitives ImageCollection"
    )
    
    parser.add_argument(
    "-o",
    "--output",
    type=str,
    required=True,
    help="GEE asset path for export"
    )

    parser.add_argument(
    "-d",
    "--dry_run",
    dest="dry_run",
    action="store_true",
    help="goes through checks and prints output asset path but does not export.",
    )
    args = parser.parse_args()

    input_path = args.input
    output_path = args.output
    dry_run = args.dry_run
    
    # If output Image exists already, throw error
    assert check_exists(output_path), f"Output image already exists: {output_path}"
    
    # If input ImageCollection does not exist, throw error
    assert check_exists(input_path) == 0, f"Input Primitives Collection does not exist: {input_path}"
    
    if dry_run:
            print(f"would export: {output_path}")
    else:
      prims = Primitives(asset_id=input_path)
      max = prims.assemble_max_probability()
      aoi = prims.collection.first().geometry().bounds()
      description = os.path.basename(output_path).replace('/','_')
      export_img_to_asset(image=max,
                          description=description,
                          assetId=output_path,
                          region=aoi)

if __name__ == "__main__":
   main()    


import ee
import hydrafloods as hf
from rlcms.harmonics import doHarmonicsFromOptions
from rlcms.covariates import indices
from rlcms.covariates import returnCovariatesFromOptions
from ee.ee_exception import EEException

ee.Initialize()

idx = indices()

def get_agg_timing(collection:hf.Dataset,**kwargs):
    """utility function for hf.Dataset.aggregate_time(). Formats `period`, `period_unit`, and `dates` args
        to create certain types of composites (defined by `composite_mode`)
    args:
        collection (hf.Dataset): Hydrafloods Dataset
    kwargs:
        composite_mode (str): one of 'annual' or 'seasonal', Default = 'annual'
        season (list[str|int]): consecutive list of months (e.g. ['01','02','03']) comprising the season.
            A required arg if composite_mode == 'seasonal'
    Returns:
        tuple(period(int),period_unit(str),dates(list[str]))
            
    """
    if 'composite_mode' not in kwargs:
        composite_mode = 'annual'
    else:
        composite_mode = kwargs['composite_mode']
    
    # get unique yyyy strings
    years = list(set([d.split(' ')[0].split('-')[0] for d in collection.dates])) 
    years.sort()
    
    if composite_mode == 'seasonal':
        if 'season' not in kwargs:
            raise ValueError("season arg required if composite_mode == 'seasonal'")
        else:
            season = kwargs['season'] # must be consecutive 
            period = len(season)
            period_unit = 'month'
            dates = [f"{y}-{str(season[0])}-01" for y in years]
    elif composite_mode == 'annual':
        period = 1
        period_unit = 'year'
        dates = [y+'-01-01' for y in years]
    
    else:
        raise ValueError(f"{composite_mode} not a valid 'composite_mode'. Choose one of 'annual' or 'seasonal'")
    
    return period,period_unit,dates

class Composite:
    """Initializes Composite class
    
        Processes multi-band composite of your chosen dataset(s) within an AOI footprint polygon

        args:
            dataset (str): one of: 'Landsat5','Landsat7','Landsat8','Sentinel1Asc','Sentinel1Desc','Sentinel2','Modis','Viirs')
            region (ee.FeatureCollection): area of interest
            start_date (str): start date
            end_date (str): end date
        
        kwargs:
            indices:list[str]
            composite_mode:str One of ['seasonal','annual'] Default = 'annual' 
            season:list[str|int]
            reducer:str|ee.Reducer
            addTasselCap:bool
            addTopography:bool
            addJRC:bool
            harmonicsOptions:dict in this format: {'nir':{'start':int[1:365],'end':[1:365]}}
        
        returns:
            ee.Image: multi-band image composite within region
        """
    
    def __init__(self,dataset:str,
                    region:ee.FeatureCollection,
                    start_date:str,
                    end_date:str,
                    **kwargs):
    
        self.dataset=dataset
        if isinstance(region, ee.Geometry):
            self.region=region.getInfo()['coordinates']
        elif isinstance(region, ee.FeatureCollection):
            self.region = region.geometry().getInfo()['coordinates']
        self.start_date=start_date
        self.end_date=end_date
        
        for key, value in kwargs.items():
            setattr(self, key, value)
        
        # testing whether need to go b/w FC and Geometry for multi_poly
        if isinstance(region,ee.FeatureCollection):
            region = region.geometry()
            region_fc = region
        elif isinstance(region,ee.Geometry):
            region = region
        else:
            raise TypeError(f"{region} must be of type ee.FeatureCollection or ee.Geometry, got {type(region)}")
        # all hydrafloods.Dataset sub-classes
        ds_dict = {'Landsat5':hf.Landsat5(region,start_date,end_date),
                'Landsat7':hf.Landsat7(region,start_date,end_date),
                'Landsat8':hf.Landsat8(region,start_date,end_date),
                'Landsat9':hf.Landsat9(region,start_date,end_date),
                'Sentinel1':hf.Sentinel1(region,start_date,end_date),
                'Sentinel1Asc':hf.Sentinel1Asc(region,start_date,end_date),
                'Sentinel1Desc':hf.Sentinel1Desc(region,start_date,end_date),
                'Sentinel2':hf.Sentinel2(region,start_date,end_date),
                'MODIS':hf.Modis(region,start_date,end_date),
                'VIIRS':hf.Viirs(region,start_date,end_date)}
        
        # dataset can either be a named dataset string supported by a hf.Dataset sub-class 
        # or a GEE Asset path
        if isinstance(dataset,str):
            if dataset in ds_dict.keys(): 
                ds = ds_dict[dataset]
            else:
                if '/' in dataset: 
                    try:
                        ds = hf.Dataset(asset_id=dataset,region=region,start_time=start_date,end_time=end_date)
                    except:
                        raise EEException
                else: 
                    raise ValueError(f"Could not construct a hf.Dataset from dataset name provided: {dataset}")
        else:
            raise TypeError(f"dataset must be str type, got: {type(dataset)}")
        
        # mask imgs to geometries in multi_poly mode
        if 'multi_poly' in kwargs:
            if kwargs['multi_poly'] == True:
                def update_mask(img):
                    ref_poly_img = ee.Image(1).paint(region_fc).Not().selfMask() # aoi can be ee.Geometry or ee.FeatureCollection for this
                    return ee.Image(img).updateMask(ref_poly_img)
                # do we want to warn user against using multi_poly unnecessarily if aoi is a single geometry?
                # would require another synchronous request of aoi's type/element size
                ds = ds.apply_func(update_mask)
        
        ds = ds.apply_func(returnCovariatesFromOptions,**kwargs)
        
        # set reducer passed to aggregate_time(), default mean
        if 'reducer' in kwargs:
            reducer=kwargs['reducer']
        else:
            reducer = 'mean'
        
        period,period_unit,dates = get_agg_timing(ds,**kwargs)
        
        # aggregate hf.Dataset
        agg_time_result = (ds.aggregate_time(reducer=reducer,
                                        rename=False,
                                        period_unit=period_unit,
                                        period=period,
                                        dates=dates)
                                        )
        
        composite = ee.ImageCollection(agg_time_result.collection).toBands()
        
        # rename bands depending on number of resulting images
        if agg_time_result.n_images > 1:
            bnames = composite.bandNames().map(lambda b: ee.String('t').cat(b))
        else:
            bnames = composite.bandNames().map(lambda b: ee.String(b).slice(2))
        
        composite = composite.rename(bnames)
            
        # compute harmonics if desired
        if 'harmonicsOptions' in kwargs:
            harmonics_features = doHarmonicsFromOptions(ds.collection,**kwargs) # returns an ee.Image, not a hf.Dataset
            composite = composite.addBands(harmonics_features)
        
        # add JRC variables if desired
        if 'addJRCWater' in kwargs:
            if kwargs['addJRCWater']:
                composite = idx.addJRC(composite).unmask(0)
        
        # add topography variables if desired     
        if 'addTopography' in kwargs:
            if kwargs['addTopography']:
                composite = idx.addTopography(composite).unmask(0)
        
        self.bands = composite.bandNames().getInfo()
        self.image = (composite.clip(region).set('dataset',dataset,
                                                     'start',start_date,
                                                     'end',end_date)
                                                    .set(kwargs)
                                                    )
        return

def stack(composites:list):
    """
    stacks a list of rlcms.Composites together, prefixing each band with the Composite's dataset name
    args:
        composites: list[rlcms.Composite]
    returns:
        ee.Image
    """
    if len(composites) < 2 or not all(isinstance(c,Composite) for c in composites):
        raise ValueError("composites must be a list of 2 or more rlcms.Composites")
    else:
        # returns list of ee.Images with renamed bands
        renamed = [c.image.regexpRename('^', f"{c.dataset.replace('/','_')}_") for c in composites]
        # stack renamed image list into one ee.Image
        stacked = ee.Image.cat(renamed)
        return stacked


import ee
import os
from pathlib import Path
import argparse
from rlcms.utils import check_exists
from rlcms.primitives import Primitives

def main():
    ee.Initialize()

    parser = argparse.ArgumentParser(
    description="Create Land Cover Primitives For All Classes in Provided Training Data",
    usage = "primitives -i path/to/input_stack -t path/to/training_data -c LANDCOVER -o path/to/output --metrics_folder local/folder/path"
    )
    
    parser.add_argument(
        "-i",
        "--input_stack",
        type=str,
        required=True,
        help="full asset path to input stack"
    )
    
    parser.add_argument(
    "-t",
    "--training_data",
    type=str, 
    nargs='+',
    required=True,
    help="full asset path(s) to training point dataset(s)"
    )

    parser.add_argument(
    "-c",
    "--class_name",
    type=str, 
    required=True,
    help="model class label"
    )
    
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        required=True,
        help="The output asset path of the primitives image collection."
    )
    
    parser.add_argument(
    "-crs",
    type=str,
    required=False,
    help="CRS string in format of EPSG:xxxxx. Defaults to EPSG:4326"
    )

    parser.add_argument(
    "-scale",
    type=int,
    required=False,
    help="output scale"
    )

    parser.add_argument(
        "--metrics_folder",
        type=str,
        required=False,
        help="The local folder to export metrics files."
    )
    
    parser.add_argument(
        "-d",
        "--dry_run",
        dest="dry_run",
        action="store_true",
        help="goes through checks and prints paths to outputs but does not export them.",
        )

    args = parser.parse_args()

    input_stack_path = args.input_stack
    train_paths = args.training_data
    class_name =args.class_name
    output = args.output
    crs = args.crs
    scale = args.scale
    metrics_path = args.metrics_folder
    dry_run = args.dry_run

    # Run Checks
    # Check input stack exists
    assert check_exists(input_stack_path) == 0, f"Check input_stack asset exists: {input_stack_path}"
    
    # Check training data exists
    for train_path in train_paths:
        assert check_exists(train_path) == 0, f"Check training_data asset exists: {train_path}"
             
    img_coll_path = output
    output_folder = os.path.dirname(output)
    # check output folder exists
    assert check_exists(output_folder) == 0, f"Check parent folder exists: {output_folder}"
    
    # don't allow ee.Image exports to pre-existing ee.ImageCollections
    if check_exists(img_coll_path) == 0:
        raise AssertionError(f"Primitives ImageCollection already exists: {img_coll_path}")

    # Construct local 'metrics' folder path from -o output or a default name if not provided
    # cwd = os.getcwd()
    # metrics_path = os.path.join(cwd,"metrics",os.path.basename(img_coll_path))

    # print output locations and exit
    if dry_run: 
        print(f"Would Export Primitives ImageCollection to: {img_coll_path}\n")
        print(f"Would Export Model Metrics to: {metrics_path}\n")
        exit()
    
    else:
        # make local metrics folder
        if not os.path.exists(metrics_path):
            Path(metrics_path).mkdir(parents=True)
        print(f"Metrics will be exported to: {metrics_path}")
        
        input_stack = ee.Image(input_stack_path)
        
        if len(train_paths) > 1:
            print(f'Merging training datasets together: {train_paths}\n')
            # we must construct the ee.FeatureCollection's and merge them together
            training_data = ee.FeatureCollection(train_paths[0])
            for i in train_paths[1:]:
                training_data = training_data.merge(i)
        else:
            training_data = ee.FeatureCollection(train_paths[0])
        
        # Construct Primitives
        prims = Primitives(inputs=input_stack,
                           training=training_data,
                           class_name=class_name)
        # Export as GEE ImgColl asset
        prims.export_to_asset(collection_assetId=img_coll_path,
                              crs=crs,
                              scale=scale)
        # Export model metrics
        prims.export_metrics(metrics_path=metrics_path)

if __name__=="__main__":
    main()


import ee
ee.Initialize()
 
def distanceFilter(pts,distance):
    """Filter Points within a FeatureCollection by a minimum distance threshold"""
    withinDistance = distance

    ## From the User Guide: https:#developers.google.com/earth-engine/joins_spatial
    ## add extra filter to eliminate self-matches
    distFilter = ee.Filter.And(ee.Filter.withinDistance(**{
      'distance': withinDistance,
      'leftField': '.geo',
      'rightField': '.geo', 
      'maxError': 1
    }), ee.Filter.notEquals(**{
      'leftField': 'system:index',
      'rightField': 'system:index',

    }))
    
    distSaveAll = ee.Join.saveAll(**{
                  'matchesKey': 'points',
                  'measureKey': 'distance'
    })
    # Apply the join.
    spatialJoined = distSaveAll.apply(pts, pts, distFilter)

    # Check the number of matches.
    # We're only interested if nmatches > 0.
    spatialJoined = spatialJoined.map(lambda f: f.set('nmatches', ee.List(f.get('points')).size()) )
    spatialJoined = spatialJoined.filterMetadata('nmatches', 'greater_than', 0)

    # The real matches are only half the total, because if p1.withinDistance(p2) then p2.withinDistance(p1)
    # Use some iterative logic to clean up
    def unpack(l): 
        return ee.List(l).map(lambda f: ee.Feature(f).id())

    def iterator_f(f,list):
        key = ee.Feature(f).id()
        list = ee.Algorithms.If(ee.List(list).contains(key), list, ee.List(list).cat(unpack(ee.List(f.get('points')))))
        return list
    
    ids = spatialJoined.iterate(iterator_f,ee.List([]))

    # Clean up 
    cleaned_pts = pts.filter(ee.Filter.inList('system:index', ids).Not())
    return cleaned_pts

# This func was developed with the idea that the user wants to extract 
# training or validation sample pts from an underlying ee.Image 
# using either a set of points or polygons as reference 
def strat_sample_from_reference(img:ee.Image,collection:ee.FeatureCollection,class_band:str,scale:int,crs:str,seed:int,
                              class_values:list,class_points:list):
    """
    Generates stratified random sample pts from reference polygons with all bands from input image extracted
    
    args:
      img (ee.Image): image whose bands will be extracted to the sample points
      collection (ee.FeatureCollection): reference polygons FeatureCollection
      class_band (str): property name of the reference (i.e. 'LANDCOVER')
      scale (int): resolution to sample the grid at
      seed (int): random seed
      class_values (list): unique reference labels (e.g. [1,2,3,4])
      class_points (list): number of points to sample per label (e.g. [100,200,100,200])
    returns:
      ee.FeatureCollection of sample points 
        They will contain the properties inherited from the reference polygons, 
          a 'random' property, and all bands from the image as properties.
    """
  
    # zip class_values and class_points together so they are easily accessible by map() index
    zip_value_n = ee.List(class_values).zip(ee.List(class_points))

    # done for each [class_value,class_points] pair
    def do_by_class(value_n):
        class_value = ee.List(value_n).get(0)
        n_points = ee.List(value_n).get(1)
        filtered_poly_by_class = collection.filter(ee.Filter.eq(class_band,class_value))
        
        def fromPolys(coll):
            geom = coll.geometry()
            # generate n random pts, n_points*multiplier
            multiplier = 2
            random_pts = (ee.FeatureCollection.randomPoints(geom,ee.Number(n_points).multiply(multiplier),seed,0.001)
                          .map(lambda f: f.set(class_band,class_value))) # class_band value from underlying poly geo must copy over into each point set
            # extract raster data to the points then limit to n_points requested
            rawSample_fromPts = ee.Image(img).sampleRegions(
                collection=random_pts, 
                scale=scale,
                projection=crs, 
                tileScale=16, 
                geometries=True).randomColumn().limit(n_points,'random')
            return rawSample_fromPts
        def fromPoints(coll):
            # extract band info to every pt in coll, then limit to n_points requested
            rawSample_fromPts = ee.Image(img).sampleRegions(
                collection=coll, 
                scale=scale,
                projection=crs, 
                tileScale=16, 
                geometries=True).randomColumn().limit(n_points,'random')
            return rawSample_fromPts
        
        # check if collection's geometry type is Polygon
        types = ee.List([ee.String('Polygon'),ee.String('MultiPolygon')])
        geom_type = collection.geometry().type()
        # conditional: 1 if geom_type=='Polygon' or 'MultiPolygon', else 0 - this is a little hacky not sure if there's a better way
        check_poly = types.map(lambda type: ee.String(type).match(geom_type)).flatten().size().gte(1)
        samples = ee.Algorithms.If(check_poly,fromPolys(filtered_poly_by_class),fromPoints(filtered_poly_by_class))
        # output_properties = ee.List(img.bandNames()).add(ee.String(class_band))
        
        return ee.FeatureCollection(samples)#.map(lambda f: f.select(output_properties))
    
    output_properties = ee.List(img.bandNames()).add(ee.String(class_band))
    pts_by_class = ee.FeatureCollection(ee.List(zip_value_n).map(do_by_class)).flatten().map(lambda f: f.select(output_properties))
    return pts_by_class

def strat_sample(img,class_band,region,scale,seed,n_points,class_values,class_points,ceo_format=True):
    """
    A wrapper for ee.Image.stratifiedSample() with CEO schema formatting if desired
    Note: This function has been found to be less efficient on EECUs and Memory than those defined above.
            Use the strat_sample_w_extraction for training data generation
            strat_sample_no_extraction can be used for testing data generation (predictor bands not required)
    """
    stratSample = ee.Image(img).stratifiedSample(
        numPoints=n_points,
        classBand=class_band,
        region=region,
        scale=scale, 
        seed=seed, 
        classValues=class_values,
        classPoints=class_points,
        dropNulls=True, 
        tileScale=16,
        geometries=True)
    
    if ceo_format:
        return stratSample.map(ceoClean)
    else:
        return stratSample

def split_train_test(pts,seed):
    """stratify 80/20 train and test points"""
    
    featColl = ee.FeatureCollection(pts)
    featColl = featColl.randomColumn(columnName='random', seed=seed)
    filt = ee.Filter.lt('random',0.8)
    train = featColl.filter(filt).map(lambda f: f.select(f.propertyNames().remove('random'))) # remove 'random' property
    test = featColl.filter(filt.Not()).map(lambda f: f.select(f.propertyNames().remove('random')))

    return train, test

def ceoClean(f):
        # LON,LAT,PLOTID,SAMPLEID.,
        fid = f.id()
        coords = f.geometry().coordinates()
        return f.set('LON',coords.get(0),
                    'LAT',coords.get(1),
                    'PLOTID',fid,
                    'SAMPLEID',fid)

def plot_id_global(n,feat):
    """takes an index number (n) and adds it to current PLOTID property of a feature 
            to ensure PLOTID values are globally unique (necessary for multiple sets of AOI sampling)"""
    aoi_id = ee.String(str(n))
    f = ee.Feature(feat)
    gid = aoi_id.cat('_').cat(ee.String(f.get('PLOTID')))
    f = f.set('PLOTID',gid, 'SAMPLEID', gid)
    return f

# these two funcs were developed with the idea that the user may want to 
# generate stratified random sample pts from reference polygons, their intent 
# would match one of these two:
# 1) extract band values from an underlying ee.Image (like a composite) to the sample pts (to make training data) [strat_sample_w_extraction]
# 2) don't extract band values from an underlying ee.Image (effectively a shortcut function to turn ref polys into validation points) [strat_sample_no_extraction]
# unclear as of yet if these func's add value or are a slight variant
# strat_sample_from_reference() above was developed as an extension of strat_sample_w_extraction to allow you to do this with points or polygons
# strat_sample_no_extraction() may have no utility other than potential performance improvments over strat_sample()
def strat_sample_no_extraction(collection:ee.FeatureCollection,class_band:str,scale:int,seed:int,
                               class_values:list=None,class_points:list=None):
    """
    Generates stratified random sample pts from reference polygons. Does not extract raster data to the points. 
    
    args:
      collection (ee.FeatureCollection): reference polygons FeatureCollection
      class_band (str): property name of the reference (i.e. 'LANDCOVER')
      scale (int): resolution to sample the grid at
      seed (int): random seed
      class_values (list): unique reference labels (e.g. [1,2,3,4])
      class_points (list): number of points to sample per label (e.g. [100,200,100,200])
    returns:
      ee.FeatureCollection of sample points 
        They will contain the properties inherited from the reference polygons and a 'random' property
    """
    # zip class_values and class_points together so they are easily accessible by map() index
    zip_value_n = ee.List(class_values).zip(ee.List(class_points))

    # done for each class and its desired n_points
    def do_by_class(value_n):
        class_value = ee.List(value_n).get(0)
        n_points = ee.List(value_n).get(1)
        # get polys by each class_band value at a time
        filtered_poly_by_class = collection.filter(ee.Filter.eq(class_band,class_value))
        geom = filtered_poly_by_class.geometry()
        # generate random pts, more than user specified
        random_pts = ee.FeatureCollection.randomPoints(geom,ee.Number(n_points).multiply(2),seed) 
        # filter out pts too close to each other at user specified scale
        random_pts = (ee.FeatureCollection(distanceFilter(random_pts,scale))
                      .randomColumn().limit(n_points,'random')) # shuffle, then try to return exact # pts user specified
        # set class_band as a property to the features
        random_pts = random_pts.map(lambda f: f.set(class_band,class_value))
        return ee.FeatureCollection(random_pts)

    pts_by_class = ee.FeatureCollection(ee.List(zip_value_n).map(do_by_class)).flatten()
    
    return pts_by_class
        

# working on optimized stratified sample function not using .stratifiedSample()
def strat_sample_w_extraction(img:ee.Image,collection:ee.FeatureCollection,scale:int,crs:str,class_band:str,seed:int,
                              class_values:list,class_points:list):
  """
    Generates stratified random sample pts from reference polygons with all bands from input image extracted
    
    args:
      img (ee.Image): image whose bands will be extracted to the sample points
      collection (ee.FeatureCollection): reference polygons FeatureCollection
      class_band (str): property name of the reference (i.e. 'LANDCOVER')
      scale (int): resolution to sample the grid at
      seed (int): random seed
      class_values (list): unique reference labels (e.g. [1,2,3,4])
      class_points (list): number of points to sample per label (e.g. [100,200,100,200])
    returns:
      ee.FeatureCollection of sample points 
        They will contain the properties inherited from the reference polygons, 
          a 'random' property, and all bands from the image as properties.
    """
  
  # zip class_values and class_points together so they are easily accessible by map() index
  zip_value_n = ee.List(class_values).zip(ee.List(class_points))
  
  # done for each [class_value,class_points] pair
  def do_by_class(value_n):
    class_value = ee.List(value_n).get(0)
    n_points = ee.List(value_n).get(1)
    filtered_poly_by_class = collection.filter(ee.Filter.eq(class_band,class_value))
    geom = filtered_poly_by_class.geometry()
    # generate random pts, more than user specified
    random_pts = ee.FeatureCollection.randomPoints(geom,ee.Number(n_points).multiply(2),seed) 
    # set class_band as property 
    random_pts = random_pts.map(lambda f: f.set(class_band,class_value))
    # extract raster data to the points and try to take specific n_points specified
    rawSample_fromPts = ee.Image(img).sampleRegions(
          collection=random_pts, 
          scale=scale,
          projection=crs, 
          tileScale=16, 
          geometries=True).randomColumn().limit(n_points,'random')
    
    return ee.FeatureCollection(rawSample_fromPts)
  
  
  pts_by_class = ee.FeatureCollection(ee.List(zip_value_n).map(do_by_class)).flatten()
  return pts_by_class


from rlcms import composites
import ee
import hydrafloods as hf
ee.Initialize()

# tiny test rect in CA
region = ee.Geometry.Polygon([[-120.70162384826843,36.658181801020476],
                               [-120.65218537170593,36.658181801020476],
                               [-120.65218537170593,36.68461734040718],
                               [-120.70162384826843,36.68461734040718],
                               [-120.70162384826843,36.658181801020476]])
aoi = ee.FeatureCollection("projects/wwf-sig/assets/kaza-lc/aoi/testSeshekeWater")#.geometry()
start_date = '2021-01-01'
end_date='2021-12-31'

# class TestComposites:
def test_composite_bands():
  
  # default behavior of composite() is annual mean composite
  composite = composites.composite(dataset='Sentinel2',
                                region=aoi,
                                start_date=start_date,
                                end_date=end_date)
  assert composite.bands == ['blue_mean','green_mean','red_mean',
                                            'nir_mean','swir1_mean','swir2_mean']
def test_composite_values():
  composite = composites.composite(dataset='Sentinel2',
                                region=region,
                                start_date=start_date,
                                end_date=end_date)
  points = ee.FeatureCollection.randomPoints(region,2)
  comparison_pts = composite.image.sampleRegions(collection=points,scale=10)
  
  assert comparison_pts.aggregate_array('blue_mean').getInfo() == [363.6774193548387, 655.3863636363636]

test_composite_bands()
test_composite_values()








import ee, math
from rlcms.utils import parse_settings
class indices():

	def __init__(self):
		
		
		 
		# list with functions to call for each index
		self.functionList = {"ND_blue_green" : self.ND_blue_green, \
							 "ND_blue_red" : self.ND_blue_red, \
							 "ND_blue_nir" : self.ND_blue_nir, \
							 "ND_blue_swir1" : self.ND_blue_swir1, \
							 "ND_blue_swir2" : self.ND_blue_swir2, \
							 "ND_green_red" : self.ND_green_red, \
							 "ND_green_nir" : self.ND_green_nir, \
							 "ND_green_swir1" : self.ND_green_swir1, \
							 "ND_green_swir2" : self.ND_green_swir2, \
							 "ND_red_swir1" : self.ND_red_swir1, \
							 "ND_red_swir2" : self.ND_red_swir2, \
							 "ND_nir_red" : self.ND_nir_red, \
							 "ND_nir_swir1" : self.ND_nir_swir1, \
							 "ND_nir_swir2" : self.ND_nir_swir2, \
							 "ND_swir1_swir2" : self.ND_swir1_swir2, \
							 "R_swir1_nir" : self.R_swir1_nir, \
							 "R_red_swir1" : self.R_red_swir1, \
							 "EVI" : self.EVI, \
							 "SAVI" : self.SAVI, \
							 "IBI" : self.IBI}


	def addAllTasselCapIndices(self,img): 
		""" Function to get all tasselCap indices """
		
		def getTasseledCap(img):
			"""Function to compute the Tasseled Cap transformation and return an image"""
			
			coefficients = ee.Array([
				[0.3037, 0.2793, 0.4743, 0.5585, 0.5082, 0.1863],
				[-0.2848, -0.2435, -0.5436, 0.7243, 0.0840, -0.1800],
				[0.1509, 0.1973, 0.3279, 0.3406, -0.7112, -0.4572],
				[-0.8242, 0.0849, 0.4392, -0.0580, 0.2012, -0.2768],
				[-0.3280, 0.0549, 0.1075, 0.1855, -0.4357, 0.8085],
				[0.1084, -0.9022, 0.4120, 0.0573, -0.0251, 0.0238]
			])
		
			bands=ee.List(['blue','green','red','nir','swir1','swir2'])
			
			# Make an Array Image, with a 1-D Array per pixel.
			arrayImage1D = img.select(bands).toArray()
		
			# Make an Array Image with a 2-D Array per pixel, 6x1.
			arrayImage2D = arrayImage1D.toArray(1)
		
			componentsImage = ee.Image(coefficients).matrixMultiply(arrayImage2D).arrayProject([0]).arrayFlatten([['brightness', 'greenness', 'wetness', 'fourth', 'fifth', 'sixth']]).float()
	  
			# Get a multi-band image with TC-named bands.
			return img.addBands(componentsImage);	
			
			
		def addTCAngles(img):

			""" Function to add Tasseled Cap angles and distances to an image. Assumes image has bands: 'brightness', 'greenness', and 'wetness'."""
			
			# Select brightness, greenness, and wetness bands	
			brightness = img.select('brightness')
			greenness = img.select('greenness')
			wetness = img.select('wetness')
	  
			# Calculate Tasseled Cap angles and distances
			tcAngleBG = brightness.atan2(greenness).divide(math.pi).rename(['tcAngleBG'])
			tcAngleGW = greenness.atan2(wetness).divide(math.pi).rename(['tcAngleGW'])
			tcAngleBW = brightness.atan2(wetness).divide(math.pi).rename(['tcAngleBW'])
			tcDistBG = brightness.hypot(greenness).rename(['tcDistBG'])
			tcDistGW = greenness.hypot(wetness).rename(['tcDistGW'])
			tcDistBW = brightness.hypot(wetness).rename(['tcDistBW'])
			img = img.addBands(tcAngleBG).addBands(tcAngleGW).addBands(tcAngleBW).addBands(tcDistBG).addBands(tcDistGW).addBands(tcDistBW)
			
			return img
	
		img = getTasseledCap(img)
		img = addTCAngles(img)
		return img

	def ND_blue_green(self,img):
		img = img.addBands(img.normalizedDifference(['blue','green']).rename(['ND_blue_green']))
		return img
	
	def ND_blue_red(self,img):
		img = img.addBands(img.normalizedDifference(['blue','red']).rename(['ND_blue_red']))
		return img
	
	def ND_blue_nir(self,img):
		img = img.addBands(img.normalizedDifference(['blue','nir']).rename(['ND_blue_nir']))
		return img
	
	def ND_blue_swir1(self,img):
		img = img.addBands(img.normalizedDifference(['blue','swir1']).rename(['ND_blue_swir1']))
		return img
	
	def ND_blue_swir2(self,img):
		img = img.addBands(img.normalizedDifference(['blue','swir2']).rename(['ND_blue_swir2']))
		return img

	def ND_green_red(self,img):
		img = img.addBands(img.normalizedDifference(['green','red']).rename(['ND_green_red']))
		return img
	
	def ND_green_nir(self,img):
		img = img.addBands(img.normalizedDifference(['green','nir']).rename(['ND_green_nir']))  # NDWBI
		return img
	
	def ND_green_swir1(self,img):
		img = img.addBands(img.normalizedDifference(['green','swir1']).rename(['ND_green_swir1']))  # NDSI, MNDWI
		return img
	
	def ND_green_swir2(self,img):
		img = img.addBands(img.normalizedDifference(['green','swir2']).rename(['ND_green_swir2']))
		return img
		
	def ND_red_swir1(self,img):
		img = img.addBands(img.normalizedDifference(['red','swir1']).rename(['ND_red_swir1']))
		return img
			
	def ND_red_swir2(self,img):
		img = img.addBands(img.normalizedDifference(['red','swir2']).rename(['ND_red_swir2']))
		return img

	def ND_nir_red(self,img):
		img = img.addBands(img.normalizedDifference(['nir','red']).rename(['ND_nir_red']))  # NDVI
		return img
	
	def ND_nir_swir1(self,img):
		img = img.addBands(img.normalizedDifference(['nir','swir1']).rename(['ND_nir_swir1']))  # NDWI, LSWI, -NDBI
		return img
	
	def ND_nir_swir2(self,img):
		img = img.addBands(img.normalizedDifference(['nir','swir2']).rename(['ND_nir_swir2'])) # NBR, MNDVI
		return img

	def ND_swir1_swir2(self,img):
		img = img.addBands(img.normalizedDifference(['swir1','swir2']).rename(['ND_swir1_swir2']))
		return img
  
	def R_swir1_nir(self,img):
		# Add ratios
		img = img.addBands(img.select('swir1').divide(img.select('nir')).rename(['R_swir1_nir']));  # ratio 5/4
		return img
			
	def R_red_swir1(self,img):
		img = img.addBands(img.select('red').divide(img.select('swir1')).rename(['R_red_swir1']));  # ratio 3/5
		return img

	def EVI(self,img):
		#Add Enhanced Vegetation Index (EVI)
		evi = img.expression(
			'2.5 * ((NIR - RED) / (NIR + 6 * RED - 7.5 * BLUE + 1))', {
			  'NIR': img.select('nir'),
			  'RED': img.select('red'),
			  'BLUE': img.select('blue')
		  }).float()
	
		img = img.addBands(evi.rename(['EVI']))

		return img
	  
	def SAVI(self,img):
		# Add Soil Adjust Vegetation Index (SAVI)
		# using L = 0.5;
		savi = img.expression(
			'(NIR - RED) * (1 + 0.5)/(NIR + RED + 0.5)', {
			  'NIR': img.select('nir'),
			  'RED': img.select('red')
		  }).float()
		img = img.addBands(savi.rename(['SAVI']))

		return img
	  
	def IBI(self,img):
		# Add Index-Based Built-Up Index (IBI)
		ibi_a = img.expression(
			'2*SWIR1/(SWIR1 + NIR)', {
			  'SWIR1': img.select('swir1'),
			  'NIR': img.select('nir')
			}).rename(['IBI_A'])
	

		ibi_b = img.expression(
			'(NIR/(NIR + RED)) + (GREEN/(GREEN + SWIR1))', {
			  'NIR': img.select('nir'),
			  'RED': img.select('red'),
			  'GREEN': img.select('green'),
			  'SWIR1': img.select('swir1')
			}).rename(['IBI_B'])
		
		ibi_a = ibi_a.addBands(ibi_b)
		ibi = ibi_a.normalizedDifference(['IBI_A','IBI_B'])
		img = img.addBands(ibi.rename(['IBI']))
		
		return img

	def addTopography(self,img): 
		"""  Function to add 30m SRTM elevation and derived slope, aspect, eastness, and 
		northness to an image. Elevation is in meters, slope is between 0 and 90 deg,
		aspect is between 0 and 359 deg. Eastness and northness are unitless and are
		between -1 and 1. """

		# Import SRTM elevation data
		elevation = ee.Image("USGS/SRTMGL1_003")
		
		# Calculate slope, aspect, and hillshade
		topo = ee.Algorithms.Terrain(elevation)
		
		# From aspect (a), calculate eastness (sin a), northness (cos a)
		deg2rad = ee.Number(math.pi).divide(180)
		aspect = topo.select(['aspect'])
		aspect_rad = aspect.multiply(deg2rad)
		eastness = aspect_rad.sin().rename(['eastness']).float()
		northness = aspect_rad.cos().rename(['northness']).float()
		
		# Add topography bands to image
		topo = topo.select(['elevation','slope','aspect']).addBands(eastness).addBands(northness)
		img = img.addBands(topo)
		return img

	def addJRC(self,img):
		""" Function to add JRC Water layers: 'occurrence', 'change_abs', 
			'change_norm', 'seasonality','transition', 'max_extent' """
		
		jrcImage = ee.Image("JRC/GSW1_0/GlobalSurfaceWater")
		
		img = img.addBands(jrcImage.select(['occurrence']).rename(['occurrence']))
		img = img.addBands(jrcImage.select(['change_abs']).rename(['change_abs']))
		img = img.addBands(jrcImage.select(['change_norm']).rename(['change_norm']))
		img = img.addBands(jrcImage.select(['seasonality']).rename(['seasonality']))
		img = img.addBands(jrcImage.select(['transition']).rename(['transition']))
		img = img.addBands(jrcImage.select(['max_extent']).rename(['max_extent']))
		
		return img


	def getIndices(self,img,covariates):	
		""" add indices to image"""
		# self = indices()
		# no need to add indices that are already there
		# see TODO below, can't use removeDuplicates in .map()
		# indices = self.removeDuplicates(covariates,img.bandNames().getInfo())
		indices = covariates
		
		for item in indices:
			img = self.functionList[item](img)

		return img

	def removeDuplicates(self,covariateList,bands):
		""" function to remove duplicates, i.e. existing bands do not need to be calculated """
		# TODO: this does not scale to being mappable server side (can't use getInfo in mapped functions)
		# would need to EEify this logic to use with ee.List()'s
		return [elem for elem in covariateList if elem not in bands]

	def renameBands(self,image,prefix):
		""" renames bands with prefix """
		
		bandnames = image.bandNames()

		def mapBands(band):
			band = ee.String(prefix).cat('_').cat(band)
			return band
			
		bandnames = bandnames.map(mapBands)
		
		image = image.rename(bandnames)

		return image


def returnCovariates(img):
	"""Workflow for computing Landsat and covariates. bands and covariates are hardcoded inside the function."""
	# hard coded for now
	bands = ['blue','green','red','nir','swir1', 'swir2']	
	bandLow = ['p20_blue','p20_green','p20_red','p20_nir','p20_swir1', 'p20_swir2']
	bandHigh = ['p80_blue','p80_green','p80_red','p80_nir','p80_swir1', 'p80_swir2']

	"""Calculate the urban, builtup cropland rice and barren primitives """
	covariates = ["ND_blue_green","ND_blue_red","ND_blue_nir","ND_blue_swir1","ND_blue_swir2", \
				  "ND_green_red","ND_green_nir","ND_green_swir1","ND_green_swir2","ND_red_swir1",\
				  "ND_red_swir2","ND_nir_red","ND_nir_swir1","ND_nir_swir2","ND_swir1_swir2",\
				  "R_swir1_nir","R_red_swir1","EVI","SAVI","IBI"]
		
	index = indices()
	
	def scaleLandsat(img):
		"""Landast is scaled by factor 0.0001 """
		thermalBand = ee.List(['thermal'])
		thermal = ee.Image(img).select(thermalBand).divide(10)
					
		otherBands = ee.Image(img).bandNames().removeAll(thermalBand)
		
		scaled = ee.Image(img).select(otherBands).multiply(0.0001)
		image = ee.Image(scaled.addBands(thermal))        
	    
		return ee.Image(image.copyProperties(img))

	img = scaleLandsat(img)
		
	def addIndices(img,prefix):
		img = index.addAllTasselCapIndices(img)
		img = index.getIndices(img,covariates)
		img = index.addJRC(img).unmask(0)
		img = index.addTopography(img).unmask(0)	
		if len(prefix) > 0:	
			img = index.renameBands(img,prefix)
		return img
			
		
	down = addIndices(img.select(bandLow,bands),"p20")
	middle = addIndices(img.select(bands),"")
	up = addIndices(img.select(bandHigh,bands),"p80")
		
	img = down.addBands(middle).addBands(up)
	
	return img

# TODO: can't figure out how to detangle getIndices() and addAllTasselCapIndices() 
# so that they can be directly passed as func to hf.Dataset.apply_func()
# so we have this function that takes unnamed kwargs, passed to apply_func()..
def returnCovariatesFromOptions(img,**kwargs):
	"""
	Computes and adds image covariates according to user settings
	args:
		img (ee.Image): image to compute covariates 
		kwargs (dict): a settings dictionary 
	returns:
		img (ee.Image): multi-band image with all desired covariates
	"""
	settings = kwargs
	if 'indices' in kwargs.keys():
		if len(kwargs['indices']) > 0:
			covariates = settings['indices']
			index = indices()
		
		img = ee.Image(img)
		img = index.getIndices(img,covariates)
	
	if 'addTasselCap' in kwargs.keys():
		if kwargs['addTasselCap']:
			img = index.addAllTasselCapIndices(img)
	
	return img


import ee
import math

ee.Initialize()

def addHarmonicTerms(image):
    """add Time bands to image"""
    timeRadians = image.select("t").multiply(2 * math.pi)
    return image.addBands(timeRadians.cos().rename("cos")).addBands(
        timeRadians.sin().rename("sin")
    )

def calculateHarmonic(imageCollection: ee.ImageCollection, dependent: ee.String):
    """
    Calculate harmonic coefficients (phase and amplitude) off of an ImageCollection
    dependent: band that you fit the harmonic model for, must be contained in ImageCollection
    """
    harmonicIndependents = ee.List(["constant", "t", "cos", "sin"])
    #  Add harmonic terms as new image bands.
    harmonicLandsat = imageCollection.map(addHarmonicTerms)
    # The output of the regression reduction is a 4x1 array image.
    harmonicTrend = harmonicLandsat.select(harmonicIndependents.add(dependent)).reduce(
        ee.Reducer.linearRegression(harmonicIndependents.length(), 1)
    )

    # Turn the array image into a multi-band image of coefficients.
    harmonicTrendCoefficients = (
        harmonicTrend.select("coefficients")
        .arrayProject([0])
        .arrayFlatten([harmonicIndependents])
    )

    # // Compute phase and amplitude.
    phase = (
        harmonicTrendCoefficients.select("cos")
        .atan2(harmonicTrendCoefficients.select("sin"))
        .rename(dependent.cat("_phase"))
    )

    amplitude = (
        harmonicTrendCoefficients.select("cos")
        .hypot(harmonicTrendCoefficients.select("sin"))
        .rename(dependent.cat("_amplitude"))
    )
    return ee.Image.cat(phase, amplitude)


def harmonicRGB(harmonics: ee.Image):
    """Use the HSV to RGB transform to display phase and amplitude"""
    amplitude = harmonics.select(".*amplitude")
    phase = harmonics.select(".*phase")

    rgb = (
        phase.unitScale(-math.pi, math.pi)
        .addBands(amplitude.multiply(2.5))
        .addBands(ee.Image(1))
        .hsvToRgb()
    )
    return rgb


def addTimeConstant(imageCollection: ee.ImageCollection, timeField: str):
    """
    Add time constant to images in an ImageCollection
    timeField: time stamp property name (typically is the 'system:time_start' property of an image)
    """
    def _(image, timeField):
        # // Compute time in fractional years since the epoch.
        date = ee.Date(image.get(timeField))
        years = date.difference(ee.Date("1970-01-01"), "year")
        # // Return the image with the added bands.
        return image.addBands(ee.Image(years).rename("t").float()).addBands(
            ee.Image.constant(1)
        )

    return imageCollection.map(lambda i: _(i, timeField))

# TODO: consider using hf.timeseries module instead,
#  but doesn't look like the methods are exact same as defined here
#   ht.timeseries functions don't return phase and amplitude..
def doHarmonicsFromOptions(imgColl:ee.ImageCollection,**kwargs):
    """
    calculateHarmonic function band-wise
    
    args:
        imgColl (ee.ImageCollection)
    kwargs:
        harmonicsOptions (dict): which band(s) and the DOY start and end date to compute harmonics on
            formatted like: 
            {
                'red':{'start':1,'end':365},
                'blue':{'start':1,'end':365}
                }
    returns:
        ee.Image containing bands [band_phase, band_amplitude]
    """
    imgColl = ee.ImageCollection(imgColl)
    harmonicsOptions = kwargs['harmonicsOptions']

    # get harmonicsOptions dictionary
    if isinstance(harmonicsOptions,dict):
    
        # get band keys as list
        bands = ee.Dictionary(harmonicsOptions).keys()
        
        def harmonicByBand(band):
            band = ee.String(band)
            # get the params for that band
            bandwiseParams = ee.Dictionary(harmonicsOptions).get(band)
            
            # get the start and end DOY parameters
            start = ee.Dictionary(bandwiseParams).get('start')
            end = ee.Dictionary(bandwiseParams).get('end')
            
            # create temporal filtered imgColl for that band
            imgCollByBand = (ee.ImageCollection(imgColl)
                                .select(band)
                                .filter(ee.Filter.dayOfYear(start,end)))
            # add time bands
            timeField = "system:time_start"
            timeCollection = addTimeConstant(imgCollByBand, timeField)
        
            return ee.Image(calculateHarmonic(timeCollection,band))
    else:
        raise TypeError(f"harmonicsOptions expects dict type, got: {type(harmonicsOptions)}")
    
    # do harmonics by band key in model_inputs dictionary
    listOfImages = ee.Image.cat(ee.List(bands).map(harmonicByBand))
    bandStack = ee.Image(ee.ImageCollection.fromImages(listOfImages).toBands())
    
    # to remove srcImg band name indexing resulting from .toBands() 
    # (i.e. [0_swir1_phase, 0_swir1_amplitude] -> [swir1_phase, swir1_amplitude] )
    bandNames = bandStack.bandNames()
    fixedBandNames = bandNames.map(lambda e: ee.String(e).split("_").slice(-2).join("_"))
    return bandStack.rename(fixedBandNames)

if __name__ == "__main__":
    # inputs
    roi = ee.Geometry.Point([22.5019, -17.9789])
    imageCollection = ee.ImageCollection("COPERNICUS/S2_SR")
    timeField = "system:time_start"
    dependent = ee.String("B8")

    # Napkin tests...
    # prep image collection for test
    filteredCollection = imageCollection.filterBounds(roi).filterDate(
        "2019-01-01", "2021-01-01"
    )

    # test add time bands
    timeCollection = addTimeConstant(filteredCollection, timeField)

    assert timeCollection.select("t", "constant").first().bandNames().getInfo() == [
        "t",
        "constant",
    ]

    harmonics = calculateHarmonic(timeCollection, dependent)
    # test harmonics returns correect bands
    assert (
        harmonics.bandNames().getInfo()
        == ee.List([dependent.cat("_phase"), dependent.cat("_amplitude")]).getInfo()
    )

    rgb = harmonicRGB(harmonics)
    # test rgb returns correct bands
    eetest = ee.Algorithms.IsEqual(rgb.bandNames(), ["red", "green", "blue"]).getInfo()
    assert eetest is True

    harmonicsByOptions = doHarmonicsFromOptions(imageCollection)
    optionBands = 'swir1'
    assert (harmonicsByOptions.bandNames().getInfo() 
            == ee.List([optionBands.cat("_phase"), optionBands.cat("_amplitude")]).getInfo()
            ) 


import argparse
import os
from rlcms.utils import check_exists, exportTableToAsset
from rlcms.sampling import strat_sample_w_extraction, strat_sample_from_reference, split_train_test
import ee
import numpy as np
ee.Initialize()

def main():
    parser = argparse.ArgumentParser(
    description="Extract Train and Test Point Data from an Input Image within Reference Data Locations",
    usage = """train_test -ref path/to/reference_fc -im path/to/input/stack -band LANDCOVER --scale 10
                -o unique/output/path --class_values 1 2 3 4 5 6 7 8 --class_points 10 10 10 10 10 10 10"""
    )
    
    parser.add_argument(
    "-ref",
    "--reference_data",
    type=str,
    required=True,
    help="The full GEE asset path to the reference data"
    )
    
    parser.add_argument(
    "-im",
    "--input_img",
    type=str,
    required=True,
    help="The full asset path to the input stack ee.Image"
    )

    parser.add_argument(
    "-band",
    "--class_band",
    type=str,
    required=True,
    help="class band name to use for stratification"
        )
    
    parser.add_argument(
    "-s",
    "--scale",
    type=int,
    required=True,
    help="scale to conduct sampling (meters)"
    )
    
    parser.add_argument(
    "-c",
    "--crs",
    type=str,
    required=False,
    help="CRS string in format of EPSG:xxxxx. Defaults to EPSG:4326"
    )
    
    parser.add_argument(
    "-o",
    "--output",
    type=str,
    required=True,
    help="The output asset path basename for export"
    )
    
    parser.add_argument(
    '--class_values', 
    type=int, 
    nargs='+',
    required=True,
    help="list of unique LANDCOVER values in input Feature Collection"
    )

    parser.add_argument(
    "--class_points",
    type=int,
    nargs='+',
    required=True,
    help="number of samples to collect per class"
    )
    
    parser.add_argument(
    "-ns",
    "--no_split",
    dest="no_split",
    action="store_true",
    help="don't split extracted points into train and test",
    )
    
    parser.add_argument(
    "-r",
    "--reshuffle",
    dest="reshuffle",
    action="store_true",
    help="randomizes seed used in all functions",
    )

    parser.add_argument(
    "-d",
    "--dry_run",
    dest="dry_run",
    action="store_true",
    help="goes through checks and prints output asset path but does not export",
    )
    
    args = parser.parse_args()
    
    input_fc = args.reference_data
    input_img = args.input_img
    class_band = args.class_band
    output = args.output
    scale = args.scale
    crs = args.crs
    class_values = args.class_values
    class_points = args.class_points
    dry_run = args.dry_run
    no_split = args.no_split
    reshuffle = args.reshuffle

    # perform checks
    output_folder = os.path.dirname(output)
    
    assert check_exists(input_fc) == 0, f"Check input FeatureCollection exists: {input_fc}"
    assert check_exists(output_folder) == 0, f"Check output folder exists: {output_folder}"
    assert check_exists(input_img) == 0, f"Check input image exists: {input_img}"

    # class_values and class_points must be equal length
    if len(class_values) != len(class_points):
        raise ValueError(f"Error: class_points and class_values are of unequal length: {class_values} {class_points}")
        
    # user may not want to sample all classes, but provide warning to catch user error
    class_values_actual = ee.FeatureCollection(input_fc).aggregate_array('LANDCOVER').distinct().sort().getInfo()
    if class_values_actual != class_values:
        print(f"Warning: All classes in the reference dataset will not be sampled with class_values provided by user (EE-Reported class_values:{class_values_actual}). Processing will continue.")

    if dry_run:
        if no_split:
            print(f"would export (Asset): {output}")
            exit()
        else:
            print(f"would export (Asset): {output}_[train|test]_pts")
            exit()
    else:
        # default seed is set, will re-randomize seed if reshuffle==True
        seed=90210
        if reshuffle:
            np.random.RandomState()
            seed = np.random.randint(low=1,high=1e6)
            print(f"reshuffled new seed: {seed}")
        
        image = ee.Image(input_img)
        fc = ee.FeatureCollection(input_fc)
        pts = strat_sample_from_reference(img=image,
                                         collection=fc,
                                         class_band=class_band, 
                                         scale=scale, 
                                         crs=crs, 
                                         seed=seed,
                                         class_values=class_values,
                                         class_points=class_points)
        
        if no_split==False: # split into train and test pts
            train,test = split_train_test(pts,seed)
            train_assetid = f"{output}_train_pts"
            train_description = os.path.basename(train_assetid).replace('/','_')
            exportTableToAsset(train,train_description,train_assetid)

            test_assetid = f"{output}_test_pts"
            test_description = os.path.basename(test_assetid).replace('/','_')
            exportTableToAsset(test,test_description,test_assetid)
        
        else: # export all pts as is
            assetid = f"{output}"
            description = os.path.basename(assetid).replace('/','_')
            exportTableToAsset(pts,description,assetid)
        
if __name__ == "__main__":
    main()



